#include <Arduino.h>
#include <stdlib.h>
#include <stdint.h>
#include <config.h>
#include <DFRobot_ST7687S_Widgets.h>
#include <imgs.h>
#include "BLEDevice.h"

DFRobot_ST7687S_Widgets tft_left(TFT_LEFT_PIN_CS, TFT_LEFT_PIN_RS, TFT_LEFT_PIN_WR, TFT_LEFT_PIN_LCK);
DFRobot_ST7687S_Widgets tft_right(TFT_RIGHT_PIN_CS, TFT_RIGHT_PIN_RS, TFT_RIGHT_PIN_WR, TFT_RIGHT_PIN_LCK);

bool doConnect = false;
bool connected = false;
bool doScan = false;
int btn_checked = millis();
static BLEUUID serviceUUID("0000ffe0-0000-1000-8000-00805f9b34fb");
static BLEUUID charUUID("0000ffe1-0000-1000-8000-00805f9b34fb");
static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;

// TODO: check request length
// uint8_t status_request[] = { 0x66, 0x41, 0x02, 0x05, 0x00, 0xAE, 0x00, 0x00, 0x30, 0xDD, 0xFC, 0x3F, 0xF0, 0xDC, 0xFC, 0x3F, 0x66, 0x10, 0x94, 0xEE, 0x13, 0x00, 0x00, 0x00 };
uint8_t old_speed = 0;
uint8_t status_request[] = { 0x66, 0x41, 0x02, 0x05, 0x00, 0xAE };
uint8_t last_bike_status[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
uint8_t bike_status[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };

float calc_speed() {
  int rmp = (int) (60.0d / (((double) ((bike_status[2] * 255) + bike_status[3])) / 1000.0d));
  uint8_t tempSpeed = (int) ((((double) rmp) * ((((double) ((float) 27) / 39.37f)) * 3.141592653589793d) / 1000.0d) * 60.0d);
  return tempSpeed;
}

static void notifyCallback(BLERemoteCharacteristic* pBLERemoteCharacteristic, uint8_t* pData, size_t length, bool isNotify) {

  if ( (length == 9) && (pData[0] ==0x66) && (pData[1] ==0x41) && (pData[2] ==0x05) ) {
    for (int i = 0; i < length; i++) {
      last_bike_status[i] = bike_status[i];
      bike_status[i] = pData[i];
    }
  }

  else {
    Serial.print("size: ");
    Serial.print(length);
    Serial.print(" | data: ");
    for (int i = 0; i < length; i++) {
      Serial.print(pData[i], HEX);
      Serial.print(" ");
    }
    Serial.println("");
  }
}

class MyClientCallback : public BLEClientCallbacks {
  void onConnect(BLEClient* pclient) { }
  void onDisconnect(BLEClient* pclient) { connected = false; }
};

class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
  void onResult(BLEAdvertisedDevice advertisedDevice) {
    if (advertisedDevice.getAddress().equals(BLEAddress(RAYVOLT_BIKE))) {
      BLEDevice::getScan()->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;
    }
    if (advertisedDevice.getAddress().equals(BLEAddress(RAYVOLT_DUMMY))) {
      BLEDevice::getScan()->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;
    }
  }
};

bool connectToServer() {
    BLEClient*  pClient  = BLEDevice::createClient();
    pClient->setClientCallbacks(new MyClientCallback());
    pClient->connect(myDevice);
    BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
    if (pRemoteService == nullptr) {
      pClient->disconnect();
      return false;
    }

    pRemoteCharacteristic = pRemoteService->getCharacteristic(charUUID);
    if (pRemoteCharacteristic == nullptr) {
      pClient->disconnect();
      return false;
    }

    if(pRemoteCharacteristic->canRead()) {
      std::string value = pRemoteCharacteristic->readValue();
    }

    if(pRemoteCharacteristic->canNotify())
      pRemoteCharacteristic->registerForNotify(notifyCallback);

    connected = true;
    return true;
}

void process_status() {
//  TODO: ERRORS OverVoltage / Undervoltage / BLE_ERROR
/*
  Serial.print("bike_status");
  for (int i = 0; i < sizeof(bike_status); i++) {
    Serial.print(bike_status[i], BIN);
    Serial.print(" ");
  }
  Serial.println("");
*/
  if( ((last_bike_status[7] >> 0) & 1) != ((bike_status[7] >> 0) & 1) ) {
    if((bike_status[7] >> 0) & 1) {
      Serial.println("rbs error");
      tft_right.drawBmp(RegenBrakeFaultBMP, -30, 30, 25, 25, 1, REGENBRAKE_FAULT_COLOR);
    }
    else {
      tft_right.fillRect(-30, 30, 25, 25, CBC_DISPLAY_BACKCOLOR);
    }
  }
  
  if( ((last_bike_status[7] >> 1) & 1) != ((bike_status[7] >> 1) & 1) ) {
    if((bike_status[7] >> 1) & 1) {
     tft_right.drawBmp(ThrottleFaultBMP, 5, 30, 25, 25, 1, THROTTLE_FAULT_COLOR);
    }
    else {
      tft_right.fillRect(5, 30, 25, 25, CBC_DISPLAY_BACKCOLOR);
    }
  }
}

void setup() {
  Serial.begin(115200);

  pinMode(BTN_RIGHT_LOWER, INPUT);
  pinMode(BTN_RIGHT_CENTER, INPUT);
  pinMode(BTN_RIGHT_HIGHER, INPUT);

  tft_left.begin(ST7687S_Rotation_270);
  tft_left.fillScreen(CBC_DISPLAY_BACKCOLOR);

  tft_right.begin(ST7687S_Rotation_90);
  tft_right.fillScreen(CBC_DISPLAY_BACKCOLOR);

  tft_right.AmpMeter(50);
  tft_right.PasDisplay((status_request[3]) - 2);
//  tft_right.fillRect(-30, 30, 25, 25, DISPLAY_GREEN);
//  tft_right.drawBmp(RegenBrakeFaultBMP, -30, 30, 25, 25, 1, DISPLAY_ORANGE);
//  tft_right.fillRect(5, 30, 25, 25, DISPLAY_GREEN);
//  tft_right.drawBmp(ThrottleFaultBMP, 5, 30, 25, 25, 1, DISPLAY_ORANGE);
  tft_right.drawBmp(Battery5BMP, 30, -28, 25, 50, 1, DISPLAY_GREEN);
  tft_left.Tacho();
  tft_left.TachoLabels();
  tft_left.TachoNeedle(35);

//  tft_left.drawLine2(0 ,0, -45, -45, DISPLAY_DARKGREY);
  BLEDevice::init("");
  BLEScan* pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(1349);
  pBLEScan->setWindow(449);
  pBLEScan->setActiveScan(true);
  pBLEScan->start(5, false);
}

void loop() {
  if (touchRead(TOUCH_BUTTON) < TOUCH_THRESHOLD) {
    Serial.println("touch detected");
  }

  if (doConnect == true) {
    if (connectToServer()) { Serial.println("connected"); }
    else { Serial.println("failed to connect"); }
    doConnect = false;
  }

  if (connected) {
    pRemoteCharacteristic->writeValue(status_request, sizeof(status_request));

    if ( (digitalRead(BTN_RIGHT_LOWER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
      btn_checked = millis();

      if (status_request[3] > 2 ) {
        status_request[3]--;
        tft_right.PasDisplay((status_request[3]) - 2);
      }
    }

    else if ( (digitalRead(BTN_RIGHT_CENTER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
      btn_checked = millis();
    }

    else if ( (digitalRead(BTN_RIGHT_HIGHER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
      btn_checked = millis();

      if (status_request[3] < 5 ) {
        status_request[3]++;
        tft_right.PasDisplay((status_request[3]) - 2);
      }
    }
    process_status();
  }

  else if(doScan){
    Serial.println("disconnected");
//    BLEDevice::getScan()->start(0);  // this is just eample to start scan after disconnect, most likely there is better way to do it in arduino
  }

  delay(250);
}